"
 (C) 2010 by Holger Hans Peter Freyther
 All Rights Reserved

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"

Object subclass: SCCPHelper [
    <category: 'osmo-network'>
    SCCPHelper class >> msgCr   [ ^ 16r01 ]
    SCCPHelper class >> msgCc   [ ^ 16r02 ]
    SCCPHelper class >> msgCref [ ^ 16r03 ]
    SCCPHelper class >> msgRlsd [ ^ 16r04 ]
    SCCPHelper class >> msgRlc  [ ^ 16r05 ]
    SCCPHelper class >> msgDt1  [ ^ 16r06 ]
    SCCPHelper class >> msgDt2  [ ^ 16r07 ]
    SCCPHelper class >> msgAk   [ ^ 16r08 ]
    SCCPHelper class >> msgUdt  [ ^ 16r09 ]
    SCCPHelper class >> msgUdts [ ^ 16r0A ]
    SCCPHelper class >> msgEd   [ ^ 16r0B ]
    SCCPHelper class >> msgEa   [ ^ 16r0C ]
    SCCPHelper class >> msgRsr  [ ^ 16r0D ]
    SCCPHelper class >> msgRsc  [ ^ 16r0E ]
    SCCPHelper class >> msgErr  [ ^ 16r0F ]
    SCCPHelper class >> msgIt   [ ^ 16r10 ]
    SCCPHelper class >> msgXudt [ ^ 16r11 ]
    SCCPHelper class >> msgXudts[ ^ 16r12 ]
    SCCPHelper class >> msgLudt [ ^ 16r13 ]
    SCCPHelper class >> msgLudts[ ^ 16r14 ]

    SCCPHelper class >> pncData [ ^ 16r0F ]
    SCCPHelper class >> pncEoO  [ ^ 16r00 ]

    SCCPHelper class >> createCR: src dest: dest data: aData [
        ^ (SCCPConnectionRequest initWith: src dest: dest data: aData)
            toMessage.
    ]

    SCCPHelper class >> createRLSD: src dest: dest cause: cause [
        ^ (SCCPConnectionReleased initWith: src dest: dest cause: cause)
            toMessage.
    ]

    SCCPHelper class >> createDT1: dst data: data [
        ^ (SCCPConnectionData initWith: dst data: data)
            toMessage.
    ]
]

Object subclass: SCCPPNC [
    | dict |

    SCCPPNC class >> parseFrom: aPnc [
        | dict pnc |

        pnc := aPnc.
        dict := Dictionary new.

        [pnc isEmpty not] whileTrue: [
            | type |
            type := pnc at: 1.

            type = SCCPHelper pncEoO
                ifTrue: [
                    pnc := ByteArray new.
                ]
                ifFalse: [
                    | size data |
                    size := pnc at: 2.
                    data := pnc copyFrom: 3 to: 3 + size - 1.
                    pnc := pnc copyFrom: 3 + size.
                    dict at: type put: data.
                ].
        ].

        ^ (self new)
                dict: dict;
                yourself.
    ]

    at: aKey put: aValue [
        self dict at: aKey put: aValue.
    ]

    at: aKey [
        ^ self dict at: aKey.
    ]

    dict [
        <category: 'accessing'>
        ^ dict ifNil: [dict := Dictionary new.]
    ]

    dict: aDict [
        <category: 'private'>
        dict := aDict.
    ]

    writeOn: aMsg [
        self dict keysAndValuesDo: [:key :val |
            | dat |
            dat := val toMessageOrByteArray.

            aMsg putByte: key.
            aMsg putByte: dat size.
            aMsg putByteArray: dat.
        ].

        aMsg putByte: SCCPHelper pncEoO.
    ]
]

Object subclass: SCCPGTI [
    | indicator nai data |
    <comment: 'I represent the Global Title of Q.713.'>

    "gti indicator in the Addr header"
    SCCPGTI class >> gtiIndNoGTI                [ ^ 0 ]
    SCCPGTI class >> gtiIndGTI                  [ ^ 1 ]
    SCCPGTI class >> gtiIndTransOnlyGTI         [ ^ 2 ]
    SCCPGTI class >> gtiIndTransNumbrPlanAndEnc [ ^ 3 ]
    SCCPGTI class >> gtiIndTransNumbrAndMore    [ ^ 4 ]

    "nai type"
    SCCPGTI class >> naiUnknown                 [ ^ 0 ]
    SCCPGTI class >> naiSubscriber              [ ^ 1 ]
    SCCPGTI class >> naiReservedNational        [ ^ 2 ]
    SCCPGTI class >> naiNationalSign            [ ^ 3 ]
    SCCPGTI class >> naiInternationalNumber     [ ^ 4 ]

    "numbering plan"
    SCCPGTI class >> npUnknown                  [ ^ 0 ]
    SCCPGTI class >> npISDN                     [ ^ 1 ]
    SCCPGTI class >> npGeneric                  [ ^ 2 ]
    SCCPGTI class >> npData                     [ ^ 3 ]
    SCCPGTI class >> npTelex                    [ ^ 4 ]
    SCCPGTI class >> npMaritime                 [ ^ 5 ]
    SCCPGTI class >> npLand                     [ ^ 6 ]
    SCCPGTI class >> npMobile                   [ ^ 7 ]

    "encoding scheme"
    SCCPGTI class >> esUnknown                  [ ^ 0 ]
    SCCPGTI class >> esBCDOdd                   [ ^ 1 ]
    SCCPGTI class >> esBCDEven                  [ ^ 2 ]
    SCCPGTI class >> esNational                 [ ^ 3 ]

    SCCPGTI class >> initWith: gti_ind data: gti [
        self allSubclassesDo: [:each |
            each subType = gti_ind
                ifTrue: [
                    ^ each initWith: gti.
            ].
        ].

        ^ self error: 'Unhandled gti indicator: %1' % {gti_ind}.
    ]

    SCCPGTI class >> map: aDigit [
        ^ (aDigit >= 0 and: [aDigit <= 9])
            ifTrue:  [ (aDigit + 48) asCharacter ]
            ifFalse: [ $N ]
    ]

    SCCPGTI class >> unmap: aChar [
        | digit |
        digit := aChar asInteger.
        ^ (digit >= 48 and: [digit <= 57])
            ifTrue:  [ digit - 48 ]
            ifFalse: [ 16rF ].
    ]

    SCCPGTI class >> parseAddr: data encoding: aEnc [
        | odd split |
        (aEnc = 1 or: [aEnc = 2]) ifFalse: [
            ^ self error: 'Only BCD number encoding supported.'
        ].

        split := OrderedCollection new.
        data do: [:each |
            split add: (self map: (each bitAnd: 16r0F)).
            split add: (self map: ((each bitAnd: 16rF0) bitShift: -4)).
        ].

        "Handle the odd case"
        aEnc = 1 ifTrue: [
            split removeLast.
        ].

        ^ split asString.
    ]

    SCCPGTI class >> formatAddr: aNumber on: data [
        | nr odd |

        nr := OrderedCollection new.
        odd := aNumber size odd.
        aNumber do: [:each |
            nr add: (self unmap: each)
        ].

        odd ifTrue: [
            nr add: 16rF.
        ].

        1 to: nr size by: 2 do: [:each|
            | low high |
            low := nr at: each.
            high := nr at: each + 1.
            data add: (low bitOr: (high bitShift: 4)).
        ].
    ]
]

SCCPGTI subclass: SCCPGTITranslation [
    | trans plan enc nature addr |
    SCCPGTITranslation class >> subType [ ^ 4 ]
    SCCPGTITranslation class >> initWith: data [
        | enc |
        enc := (data at: 2) bitAnd: 16r0F.
        ^ self new
            translation: (data at: 1);
            plan: (((data at: 2) bitAnd: 16rF0) bitShift: -4);
            encoding: enc;
            nature: ((data at: 3) bitAnd: 16r7F);
            addr: (self parseAddr: (data copyFrom: 4) encoding: enc);
            yourself
    ]

    translation [
        ^ trans ifNil: [ 0 ]
    ]

    translation: aTrans [ trans := aTrans ]

    plan [ ^ plan ]
    plan: aPlan [ plan := aPlan ]

    encoding [
        ^ enc ifNil: [
            addr size odd
                ifTrue:  [
                    1
                ]
                ifFalse: [
                    2
                ].
        ].
    ]
    encoding: aEnc [ enc := aEnc ]

    nature [ ^ nature ]
    nature: aNai [ nature := aNai ]

    addr [ ^ addr ]
    addr: anAddr [ addr := anAddr ]

    asByteArray [
        | data |
        data := OrderedCollection new.

        "write the header"
        data add: self translation.
        data add: ((plan bitShift: 4) bitOr: self encoding).
        data add: nature.

        "encode the number"
        SCCPGTI formatAddr: addr on: data.

        ^ data asByteArray
    ]
]

Object subclass: SCCPAddress [
    | route_ssn ssn poi gti gti_ind |

    SCCPAddress class >> ssnNotKnown [ ^ 0 ]
    SCCPAddress class >> ssnSCCPMgnt [ ^ 1 ]
    SCCPAddress class >> ssnITURsrvd [ ^ 2 ]
    SCCPAddress class >> ssnISUP     [ ^ 3 ]
    SCCPAddress class >> ssnOMA      [ ^ 4 ]
    SCCPAddress class >> ssnMAP      [ ^ 5 ]
    SCCPAddress class >> ssnHLR      [ ^ 6 ]
    SCCPAddress class >> ssnVLR      [ ^ 7 ]
    SCCPAddress class >> ssnMSC      [ ^ 8 ]
    SCCPAddress class >> ssnEIC      [ ^ 9 ]
    SCCPAddress class >> ssnAUC      [ ^ 10 ]
    SCCPAddress class >> ssnISUPSRV  [ ^ 11 ]
    SCCPAddress class >> ssnReserved [ ^ 12 ]
    SCCPAddress class >> ssnBroadISDN[ ^ 13 ]
    SCCPAddress class >> ssnTCTest   [ ^ 14 ]

    SCCPAddress class >> createWith: ssn [
        ^ (SCCPAddress new)
            ssn: ssn;
            routedOnSSN: true;
            yourself
    ]

    SCCPAddress class >> createWith: ssn poi: aPoi [
        ^ SCCPAddress new
            ssn: ssn;
            routedOnSSN: true;
            poi: aPoi;
            yourself
    ]

    SCCPAddress class >> parseFrom: aByteArray [
        | routed_ssn gti_ind gti len ai ssn poi dat |
        poi := nil.
        len := aByteArray at: 1.
        ai := aByteArray at: 2.

        "Copy the address"
        dat := aByteArray copyFrom: 3 to: len + 1.

        "Point Code"
        (ai bitAnd: 1) = 1
            ifTrue: [
                poi := (dat ushortAt: 1).
                dat := dat copyFrom: 3.
        ].

        "SSN"
        routed_ssn := (ai bitAnd: 16r40) = 16r40.
        ssn := dat at: 1.
        dat := dat copyFrom: 2.

        "GTI"
        gti_ind := (ai bitAnd: 16r3C) bitShift: -2.
        gti := dat copyFrom: 1.


        ^ SCCPAddress new
            ssn: ssn;
            poi: poi;
            routedOnSSN: routed_ssn;
            gti: gti indicator: gti_ind;
            yourself.
    ]

    routedOnSSN: aFlag [
        route_ssn := aFlag
    ]

    routedOnSSN [
        ^ route_ssn ifNil: [false]
    ]

    gti [
        ^ gti
    ]

    gtiInd [
        ^ gti_ind
    ]

    gti: aGti indicator: aGtiInd [
        gti := aGti.
        gti_ind := aGtiInd bitAnd: 16rF.
    ]

    gtiAsParsed [
        ^ gti_ind = 0
            ifTrue:  [nil]
            ifFalse: [SCCPGTI initWith: gti_ind data: gti].
    ]

    gtiFromAddr: aGti [
        gti_ind := aGti class subType.
        gti := aGti asByteArray.
    ]

    poi: aPoi [
        poi := aPoi.
    ]

    poi [
        ^ poi
    ]

    ssn: assn [
        ssn := assn
    ]

    ssn [
        <category: 'accessing'>
        ^ ssn.
    ]

    asByteArray [
        "Most simple address storing routine"
        | ai data |

        data := OrderedCollection new.

        "Create the Address Information"
        ai := 0.

        "SSN indicator"
        ai := ai bitOr: 2.
        self routedOnSSN ifTrue: [
            ai := ai bitOr: 64.
        ].

        "Point Code"
        poi ifNotNil: [
            ai := ai bitOr: 1.
        ].

        "GTI Indicator"
        gti_ind ifNotNil: [
            ai := ai bitOr: (gti_ind bitShift: 2).
        ].

        data add: ai.

        "POC"
        poi ifNotNil: [
            data add: ((poi bitAnd: 16r00FF) bitShift:  0).
            data add: ((poi bitAnd: 16rFF00) bitShift: -8).
        ].

        "SSN"
        data add: ssn.

        "GTI"
        gti_ind ifNotNil: [
            gti do: [:each | data add: each ].
        ].

        data addFirst: data size.

        ^ data asByteArray
    ]
]

Object subclass: SCCPAddrReference [
    <category: 'osmo-network'>
    SCCPAddrReference class >> store: anAddress on: aMsg [
        "Store the threee bytes of an sccp address on a messagebuffer"

        aMsg putByte: ((anAddress bitAnd: 16r000000FF) bitShift:  -0).
        aMsg putByte: ((anAddress bitAnd: 16r0000FF00) bitShift:  -8).
        aMsg putByte: ((anAddress bitAnd: 16r00FF0000) bitShift: -16).
    ]

    SCCPAddrReference class >> fromCData: anArray [
        "Parse from an CArray"

        | oct1 oct2 oct3 |
        oct1 := (anArray at: 0) bitShift:  0.
        oct2 := (anArray at: 1) bitShift:  8.
        oct3 := (anArray at: 2) bitShift: 16.

        ^ (oct1 bitOr: oct2) bitOr: oct3
    ]

    SCCPAddrReference class >> fromByteArray: anArray [
        "Parse from a ByteArray"
        | oct1 oct2 oct3 |
        oct1 := (anArray at: 1) bitShift:  0.
        oct2 := (anArray at: 2) bitShift:  8.
        oct3 := (anArray at: 3) bitShift: 16.

        ^ (oct1 bitOr: oct2) bitOr: oct3
    ]
]

Object subclass: SCCPMessage [
    SCCPMessage class >> decode: aByteArray [
        | type |
        type := aByteArray at: 1.

        SCCPMessage allSubclassesDo: [:each |
            each msgType = type
                ifTrue: [
                ^ each parseFrom: aByteArray.
            ]
        ].

        "raise exception"
        ^ Error signal: 'No handler for: %1' % {type}.
    ]
]

SCCPMessage subclass: SCCPConnectionRequest [
    <category: 'osmo-network'>
    | src dst pnc |

    SCCPConnectionRequest class >> msgType [
        <category: 'factory'>
        ^ SCCPHelper msgCr
    ]

    SCCPConnectionRequest class >> initWith: src dest: dest pnc: pnc [
        <category: 'construction'>
        ^ self new
            src: src dest: dest pnc: pnc;
            yourself
    ]

    SCCPConnectionRequest class >> initWith: src dest: dest data: data [
        <category: 'construction'>
        | pnc |
        pnc := SCCPPNC new.
        pnc at: SCCPHelper pncData put: data.

        ^ self new
            src: src dest: dest pnc: pnc;
            yourself
    ]

    SCCPConnectionRequest class >> parseFrom: aMsg [
        | src addr proto variable optional pnc |
        src := SCCPAddrReference fromByteArray: (aMsg copyFrom: 2 to: 4).
        proto := (aMsg at: 5) asInteger.
        variable := (aMsg at: 6) asInteger.
        optional := (aMsg at: 7) asInteger.

        "some sanity check"
        proto ~= 2 ifTrue: [
            Exception signal: 'Proto should be two was ', proto asString.
        ].

        "parse the address"
        addr := SCCPAddress parseFrom: (aMsg copyFrom: (6 + variable)).

        "parse the optional data"
        pnc := SCCPPNC parseFrom: (aMsg copyFrom: (7 + optional)).

        ^ SCCPConnectionRequest initWith: src dest: addr pnc: pnc.
    ]

    src [
        <category: 'accessing'>
        ^ src
    ]

    dest [
        <category: 'accessing'>
        ^ dst
    ]

    data [
        <category: 'accessing'>
        ^ pnc at: SCCPHelper pncData.
    ]

    data: aData [
        pnc at: SCCPHelper pncData put: aData.
    ]

    src: aSrc dest: aDest pnc: aPnc [
        src := aSrc.
        dst := aDest.
        pnc := aPnc.
    ]

    writeOn: aMsg [
        <category: 'conversion'>
        | dat len addr |

        addr := dst asByteArray.

        aMsg putByte: self class msgType.
        SCCPAddrReference store: src on: aMsg.
        "store proto_class, variable_called, optional_start"
        aMsg putByte: 2.
        aMsg putByte: 2.
        aMsg putByte: 1 + addr size.

        aMsg putByteArray: addr.

        " place the data now "
        pnc writeOn: aMsg.
        ^ aMsg.
    ]
]

SCCPMessage subclass: SCCPConnectionConfirm [
    | src dst pnc |
    SCCPConnectionConfirm class >> msgType [
        <category: 'factory'>
        ^ SCCPHelper msgCc
    ]

    SCCPConnectionConfirm class >> initWithSrc: aSrc dst: aDst [
        ^ self new
            src: aSrc dst: aDst;
            yourself
    ]

    SCCPConnectionConfirm class >> parseFrom: aMsg [
        | src dst proto optional |
        dst := SCCPAddrReference fromByteArray: (aMsg copyFrom: 2 to: 4).
        src := SCCPAddrReference fromByteArray: (aMsg copyFrom: 5 to: 7).
        proto := aMsg at: 8.
        optional := aMsg at: 9.

        "TODO: Add additional items"
        ^ self new
            src: src dst: dst;
            yourself
    ]

    writeOn: aMsg [
        <category: 'conversion'>

        aMsg putByte: SCCPHelper msgCc.
        SCCPAddrReference store: dst on: aMsg.
        SCCPAddrReference store: src on: aMsg.
        aMsg putByte: 2.
        aMsg putByte: 1.
        self pnc writeOn: aMsg.
    ]

    src: aSrc dst: aDst [
        src := aSrc.
        dst := aDst.
    ]

    src  [ ^ src  ]
    dst [ ^ dst ]
    pnc [ ^ pnc ifNil: [ pnc := SCCPPNC new. ] ]
]

SCCPMessage subclass: SCCPConnectionData [
    | dst data |

    SCCPConnectionData class >> msgType [
        <category: 'factory'>
        ^ SCCPHelper msgDt1
    ]

    SCCPConnectionData class >> initWith: dst data: data [
        ^ (self new)
                dst: dst;
                data: data;
                yourself.
    ]

    SCCPConnectionData class >> parseFrom: aByteArray [
        | more_data var_start addr size data |
        addr := SCCPAddrReference fromByteArray: (aByteArray copyFrom: 2).
        more_data := aByteArray at: 5.
        more_data = 0 ifFalse: [
            Error signal: 'Fragmented data is not supported.'.
        ].

        var_start := aByteArray at: 6.
        size := aByteArray at: 6 + var_start.
        data := aByteArray copyFrom: (6 + var_start + 1)  to: (6 + var_start + size).

        ^ SCCPConnectionData initWith: addr data: data.
    ]

    dst: aDst [
        <category: 'private'>
        dst := aDst.
    ]

    data: aData [
        <category: 'private'>
        data := aData.

        data size > 16rFF ifTrue: [
            self error: 'Data must be < 256 in size but was %1' % {data size}
        ].
    ]

    dst [
        ^ dst
    ]

    data [
        ^ data
    ]

    writeOn: aMsg [
        | dat |
        <category: 'conversion'>

        aMsg putByte: self class msgType.
        SCCPAddrReference store: dst on: aMsg.
        aMsg putByte: 0.
        aMsg putByte: 1.

        dat := data toMessageOrByteArray.
        aMsg putByte: dat size.
        aMsg putByteArray: dat.

        ^ aMsg
    ]
]

SCCPMessage subclass: SCCPConnectionReleased [
    | src dst cause pnc |

    SCCPConnectionReleased class >> msgType [
        <category: 'factory'>
        ^ SCCPHelper msgRlsd
    ]

    SCCPConnectionReleased class >> initWithDst: aDst src: aSrc cause: aCause [
        ^ self new
            dst: aDst;
            src: aSrc;
            cause: aCause;
            yourself.
    ]

    SCCPConnectionReleased class >> parseFrom: aByteArray [
        | dst src cause |

        dst := SCCPAddrReference fromByteArray: (aByteArray copyFrom: 2).
        src := SCCPAddrReference fromByteArray: (aByteArray copyFrom: 5).
        cause := aByteArray at: 8.

        ^ SCCPConnectionReleased initWithDst: dst src: src cause: cause.
    ]

    dst [ ^ dst ]
    src [ ^ src ]
    cause [ ^ cause ]
    dst: aDst [ dst := aDst ]
    src: aSrc [ src := aSrc ]
    cause: aCause [ cause := aCause ]

    pnc [
        ^ pnc ifNil: [pnc := SCCPPNC new]
    ]


    writeOn: aMsg [
        aMsg putByte: self class msgType.
        SCCPAddrReference store: dst on: aMsg.
        SCCPAddrReference store: src on: aMsg.
        aMsg putByte: cause.
        aMsg putByte: 1.
        self pnc writeOn: aMsg.
    ]
]

SCCPMessage subclass: SCCPConnectionReleaseComplete [
    | dst src |
    SCCPConnectionReleaseComplete class >> msgType [
        ^ SCCPHelper msgRlc.
    ]

    SCCPConnectionReleaseComplete class >> initWithDst: aDst src: aSrc [
        ^ self new
            dst: aDst; src: aSrc;
            yourself
    ]

    SCCPConnectionReleaseComplete class >> parseFrom: aByteArray [
        ^ self new
            dst: (SCCPAddrReference fromByteArray: (aByteArray copyFrom: 2 to: 4));
            src: (SCCPAddrReference fromByteArray: (aByteArray copyFrom: 5 to: 7));
            yourself
    ]

    dst [ ^ dst ]
    dst: aDst [ dst := aDst. ]
    src [ ^ src ]
    src: aSrc [ src := aSrc. ]

    writeOn: aMsg [
        aMsg putByte: self class msgType.
        SCCPAddrReference store: dst on: aMsg.
        SCCPAddrReference store: src on: aMsg.
    ]
]

SCCPMessage subclass: SCCPUDT [
    | called calling data error udtClass |

    SCCPUDT class >> msgType [
        ^ SCCPHelper msgUdt
    ]

    SCCPUDT class >> initWith: aCalled  calling: aCalling data: aData [
        ^ self new
            calledAddr: aCalled;
            callingAddr: aCalling;
            data: aData;
            yourself
    ]

    SCCPUDT class >> parseFrom: aByteArray [
        | called calledData calling callingData data dataData dataSize |

        called := aByteArray at: 3.
        calledData := aByteArray copyFrom: (3 + called).

        calling := aByteArray at: 4.
        callingData := aByteArray copyFrom: (4 + calling).

        data := aByteArray at: 5.
        dataSize := aByteArray at: (5 + data).
        dataData := aByteArray copyFrom: (6 + data) to: 5 + data + dataSize.

        ^ (SCCPUDT initWith: (SCCPAddress parseFrom: calledData)
                  calling:  (SCCPAddress parseFrom: callingData)
                  data: dataData)
            udtClass: ((aByteArray at: 2) bitAnd: 16r0F);
            errorHandling: ((aByteArray at: 2) bitShift: -4);
            yourself.
    ]

    calledAddr: aCalled [
        called := aCalled
    ]

    calledAddr [
        ^ called
    ]

    callingAddr: aCalling [
        calling := aCalling
    ]

    callingAddr [
        ^ calling
    ]

    data [
        ^ data
    ]

    data: aData [
        data := aData.
    ]

    errorHandling: aStrategy [
        error := aStrategy.
    ]

    errorHandling [
        ^ error ifNil: [0]
    ]

    udtClass: aClass [
        udtClass := aClass.
    ]

    udtClass [
        ^ udtClass ifNil: [0]
    ]

    writeOn: aMsg [
        | calledData callingData dat |
        calledData := called asByteArray.
        callingData := calling asByteArray.

        aMsg putByte: self class msgType.
        aMsg putByte: (((self errorHandling bitAnd: 16r0F) bitShift: 4) bitOr: self udtClass).

        "pointers"
        aMsg putByte: 3.
        aMsg putByte: 1 + calledData size + 1.
        aMsg putByte: calledData size + callingData size + 1.

        "the data"
        aMsg putByteArray: calledData.
        aMsg putByteArray: callingData.

        dat := data toMessageOrByteArray.
        aMsg putByte: dat size.
        aMsg putByteArray: dat.
    ]
]

SCCPMessage subclass: SCCPInactivityTest [
    | src dst proto seq credit |

    SCCPInactivityTest class >> msgType [
        <category: 'factory'>
        ^ SCCPHelper msgIt
    ]

    SCCPInactivityTest class >> initWithDst: aDst src: aSrc [
        <category: 'factory'>
        ^ self new
            dst: aDst;
            src: aSrc;
            yourself.
    ]

    SCCPInactivityTest class >> parseFrom: aByteArray [
        | dst src proto seq credit |
        <category: 'factory'>

        dst := SCCPAddrReference fromByteArray: (aByteArray copyFrom: 2).
        src := SCCPAddrReference fromByteArray: (aByteArray copyFrom: 5).
        ^ (self initWithDst: dst src: src)
            instVarNamed: #proto put: (aByteArray at: 8);
            instVarNamed: #seq put: (aByteArray copyFrom: 9 to: 10);
            instVarNamed: #credit put: (aByteArray at: 11);
            yourself
    ]

    src: aSrc [
        <category: 'stuff'>
        src := aSrc.
    ]

    src [
        <category: 'stuff'>
        ^ src
    ]

    dst: aDst [
        <category: 'stuff'>
        dst := aDst
    ]

    dst [
        <category: 'stuff'>
        ^ dst
    ]

    credit [
        <category: 'stuff'>
        ^ credit ifNil: [0]
    ]

    credit: aCredit [
        <category: 'stuff'>
        credit := aCredit
    ]

    protoClass [
        <category: 'stuff'>
        ^ proto ifNil: [0]
    ]

    protoClass: aClass [
        <category: 'stuff'>
        proto := aClass.
    ]

    seq [
        <category: 'stuff'>
        ^ seq ifNil: [ByteArray new: 2]
    ]

    seq: aSeq [
        <category: 'stuff'>
        seq := aSeq.
    ]

    writeOn: aMsg [
        aMsg putByte: self class msgType.
        SCCPAddrReference store: dst on: aMsg.
        SCCPAddrReference store: src on: aMsg.
        aMsg putByte: self protoClass.
        aMsg putByteArray: self seq.
        aMsg putByte: self credit.
    ]
]
